package ASCOPE::Apache::XSLT;
use strict;

$ASCOPE::Apache::XSLT::VERSION = '0.3';

use Apache;
use Apache::Constants qw (:common);
use Apache::Log;
use Apache::URI;

use XML::LibXML;
use XML::LibXSLT;

use Cache::FileCache;
use Digest::MD5 qw (md5_hex);

sub handler {
  my $apache = shift;

  # Hello, Apache::Filter

  if ($apache->dir_config("Filter") =~ /^on$/i) {
    $apache = $apache->filter_register();
  }

  # Pull in some stylesheets - if there are none
  # then are work is done, which pretty much begs
  # the question.

  my @styles = ($apache->dir_config("ASCOPE_reset_styles") =~ /on/i) ? 
      $apache->dir_config("ASCOPE_style") : $apache->dir_config->get("ASCOPE_style");

  if (! @styles) { return DECLINED; }

  my $contenttype = $apache->dir_config("ASCOPE_contenttype");
  if (! $contenttype) { return DECLINED; }

  # Pull in any parameters that may have been passed
  # with the document request
  my %params = ($apache->method() eq "POST") ? $apache->content() : $apache->args();

  # We're going to cache the stat values on the
  # source document and the stylesheets since
  # we'll have to check the mtimes a second time,
  # if one or more documents have been modified.
  my %stat = ();

  my $created  = undef;
  my $modified = undef;
  my $force    = exists({$apache->args()}->{'__force'});

  # Create the root name for all the various cache
  # files.
  my $fname = $apache->filename();

  # Tack on the CGI parameters if that's something
  # we care about.

  if ($apache->dir_config()->get("cgiparameter")) {
    # Because I can't seem to get $apache->path_info
    # to return anything :-(
    $fname .= "&".$apache->parsed_uri()->query();
  }

  # Munge it into something unique and scary to 
  # look at
  $fname = &md5_hex($fname);

  # Try to set up cache repository

  my $cache = Cache::FileCache->new({'namespace'=>__PACKAGE__,
				     'cache_root'=>($apache->dir_config("ASCOPE_cachedir") || "/tmp"),
				     'directory_umask'=>'027',});

  if (! $cache) {
    $apache->log()->error("Failed to create Cache::FileCache object, $!");
    return DECLINED;
  }

  #

  unless ($force) {

    # First thing we do is to check the source document
    # and the stylesheets to see if they changed since
    # the last cache was created.

    my $final = &md5_hex($fname.":".$styles[$#styles]);

    if (my $last = $cache->get_object($final)) {
      $created = $last->get_created_at();
      $apache->log()->debug("$final ($created)");
    }

    else { $modified = 1; }

    # Have any of the various documents been modified 
    # since the last time we visited?

    foreach my $doc ($apache->filename(),@styles) {
      last if ($modified);

      $stat{$doc} = (stat($doc))[9];

      if ($stat{$doc} > $created) {
	$apache->log()->debug("$doc has been modified ($stat{$doc} > $created)");
	$modified = 1;
      }
    }
    
    # Nothing's changed, so we'll just send the cached
    # version and exit.
    
    if (! $modified) {
      $apache->log()->debug("No modifications - sending cache");

      $apache->content_type($contenttype);
      $apache->send_http_header();
      print $cache->get($final);
      return OK;
    }
  }

  # Okay, something has changed so we've got to create
  # objects to parse and transform one or more documents.

  my $libxml  = XML::LibXML->new();
  my $libxslt = XML::LibXSLT->new();

  # Set up some stubs that we will use to create objects
  # with

  my $xmldoc     = undef;
  my $xsldoc     = undef;
  my $stylesheet = undef;
  my $results    = undef;

  # Check to see whether any of the parameters passed
  # with the document request should be handed off to
  # the XSL processor. (Eventually, it would be nice
  # to have a finer grain control to specify parameters
  # on a per-stylesheet basis.)
  my %xslparameters = ();

  # Pull in the CGI parameters first, making sure that
  # we don't pass an empty parameter to XML::LibXSLT
  # because it will freak out and die.
  foreach my $p ($apache->dir_config->get("ASCOPE_cgiparameter")) {
    next unless ($params{$p});
    $xslparameters{$p} = $params{$p};
  }

  # Pull in fixed XSL parameters
  # Note that if there are conflicting cgi parameters
  # and xsl/directory parameters, the latter will always
  # clobber the former.
  foreach my $x ($apache->dir_config->get("ASCOPE_xslparameter")) {
    $x =~ /^([^\s]+)\s(.*)$/;
    next unless ($2);

    $xslparameters{$1} = $2;
  }

  # Start transforming.

  # As soon as the '$is_modified' flag is true, we'll stop
  # doing checks and just start applying any subsequent 
  # transformations.
  my $is_modified = $force;

  # Check to see if the document that has changed is the
  # actual document, or the first stylesheet. If this is 
  # the case then we will read/parse the actual files rather
  # then the string value of the cache thingy

  if (! $is_modified) {
    $stat{$apache->filename()} ||= (stat($apache->filename()))[9];
    $is_modified = $stat{$apache->filename()} > $created;
  }

  if (! $is_modified) {
    $stat{$styles[0]} ||= (stat($styles[0]))[9];
    $is_modified = $stat{$styles[0]} > $created;
  }

  # Set up some keys so we'll now where to read from ($last_key)
  # and write to ($key)

  my $key         = $fname;
  my $last_key    = $key;

  # Parse the source doc and first stylesheet, if necesary.

  if ($is_modified) {
    $apache->log()->debug("Transforming src document.");

    eval {

      if ($apache->dir_config("Filter") =~ /^on$/i) {
	$xmldoc = $libxml->parse_fh($apache->filter_input());
      }

      else {
	$xmldoc = $libxml->parse_file($apache->filename());
      }

      #

      $xsldoc  = $libxml->parse_file($styles[0]);

      $stylesheet = $libxslt->parse_stylesheet($xsldoc);
      $results    = $stylesheet->transform($xmldoc,%xslparameters);

      $cache->set($key,$stylesheet->output_string($results));
    };

    if ($@) {
      $apache->log()->error("Failed to transform source, $@");
      return DECLINED;
    }
  }

  # Check all the subsequent stylesheets.

  foreach my $doc (@styles[1..$#styles]) {
    $stat{$doc} ||= (stat($doc))[9];

    $key = &md5_hex($fname.":".$doc);

    # Okay, this stylesheet has changed so we'll
    # need to parse the file again. Otherwise, we'll
    # just send the transformed results.

    # The document requested hasn't 
    # been modified

    if (! $is_modified) {

      # Unfortunately, I'm not sure how else to get
      # the created time for $doc using the $cache object.
      my $_cache = $cache->get_object($key);

      if (# We have actually cached $doc
	  (ref($_cache)) &&
	  # $doc has not changed since then
	  ($stat{$doc} <= $_cache->get_created_at())) {

	# This is not the droid we are looking for
	$apache->log()->debug("$doc has not been modified - moving on");

	$last_key = $key;
	next;
      }
    }

    #

    eval {
      $xmldoc = $libxml->parse_string($cache->get($last_key));
      $xsldoc = $libxml->parse_file($doc);

      $stylesheet = $libxslt->parse_stylesheet($xsldoc);
      $results    = $stylesheet->transform($xmldoc,%xslparameters);

      $cache->set($key,$stylesheet->output_string($results));
    };

    if ($@) {
      $apache->log()->error("Failed to apply $doc stylesheet, $@");
      return DECLINED;
    }

    # Something in the chain has been changed, so we apply
    # all the subsequent stylesheets and regenerate the cache 
    # files.

    if (! $is_modified) {
      $apache->log()->debug("$doc has been modified - will apply remaining styles");
      $is_modified = 1;
    }

    $last_key = $key;
  }

  # Finally, we're done.
  $apache->content_type($contenttype);
  $apache->send_http_header();

  print $cache->get($last_key);
  return OK;
}

return 1;

__END__

=head1 NAME

ASCOPE::Apache::XSLT - Aaron's mod_perl XSLT handler.

=head1 SYNOPSIS

 <Directory /path/to/some/directory>
  <FilesMatch "\.xml$">
   SetHandler		perl-script
   PerlHandler		ASCOPE::Apache::XSLT

   PerlSetVar           ASCOPE_cachedir         /path/to/cache

   PerlSetVar		ASCOPE_contenttype	text/html

   PerlAddVar		ASCOPE_style	        /path/to/make-html.xsl
   PerlAddVar		ASCOPE_style	        /path/to/do-something-with-html.xsl

   PerlAddVar           ASCOPE_xslparameter     "css_screen '/mystyles.css screen'"
  </FilesMatch>
 </Directory>

=head1 DESCRIPTION

This is Aaron's mod_perl XSLT handler. It reads a source document and 
applies one or more stylesheets,caching the output of each transformation
along the way.

B<This is not a replacement for AxKit.>

I couldn't get AxKit to do something I needed and decided that it would
probably take just as long to write my own thingy as it would to figure
out a fix to my problem.

Since I only use AxKit for doing transformations, and not any of the other
whiz-bang stuff it can do, I opted for the former.

If the handler encounters any errors during the transformation process,
it will simply return the source document as-is. (I may finesse this in
future releases.)

=head1 CONFIG DIRECTIVES

=head2 ASCOPE_contenttype

The content-type of the final output.

I<required>

=head2 ASCOPE_style

One or more XSLT stylesheets to be applied to the source
document. The output of one stylesheet is cached to disk
and fed to its subsequent stylesheet.

I<required>

=head2 ASCOPE_reset_styles

For example, you may have a series of nested directories
organized by date where you want to apply a single, unique
stylesheet for the year, month and day.

By default, the handler will apply all of the stylesheets that
are added by each I<Directory>'s B<ASCOPE_style> I<PerlSetVar>
directive.

If you enable the B<ASCOPE_reset_styles> I<PerlSetVar> directive,
then the handler will only apply the first stylesheet; that it
the most recently defined stylesheet assigned by the B<ASCOPE_style>
I<PerlSetVar> directive.

Example :

 <Directory /path/to/content>

   PerlModule		ASCOPE::Apache::XSLT

   SetHandler           perl-script
   PerlHandler          ASCOPE::Apache::XSLT

   PerlSetVar           ASCOPE_cachedir         /usr/local/www/htcache
   PerlSetVar           ASCOPE_contenttype      text/html
 </Directory>

 # Index for month

 <DirectoryMatch "/path/to/content/[0-9]{4}/[0-9]{2}">
  PerlSetVar   ASCOPE_reset_styles    On

  <FilesMatch	 "index\.html$">
   PerlSetVar   ASCOPE_style          /path/to/render-month.xsl
  </FilesMatch>

 </DirectoryMatch>

 # Index for day

 <DirectoryMatch "/path/to/content/[0-9]{4}/[0-9]{2}/[0-9]{2}">
  PerlSetVar   ASCOPE_reset_styles    On

  <FilesMatch	 "index\.html$">
   PerlSetVar   ASCOPE_style          /path/to/render-day.xsl
  </FilesMatch>

 </DirectoryMatch>

=head2 ASCOPE_cachedir

The path to the directory where cache files should be saved. If no value
is defined, then the default is '/tmp'.

If the latter, note that your cache files may be deleted whenever the 
Apache is restarted.

=head2 ASCOPE_cgiparameter

One or more CGI parameters to pass to your XSLT stylesheets.

(Eventually, it would be nice to have a finer grain control
to specify parameters on a per-stylesheet basis.)

The following list of CGI parameters are treated specially:

=over 4

=item *

B<__force>

Ignore any existing cache files and rebuilt the output,
perform all transformations.

=back

=head2 ASCOPE_xslparameter

One or more fixed parameters to pass to your XSLT stylesheets.

The parameter name is determined to be every up to, and excluding
the first white-space in the key value. The parameter value is 
everything else.

These directives should be double quoted; additionally if your parameter
value contains white-space, the entire value should be wrapped in
single quotes.

 PerlAddVar   ASCOPE_xslparameter         "cow moo"
 PerlAddVar   ASCOPE_xslparameter         "duck 'quack quack quack'"

If there ASCOPE_cgiparameter and ASCOPE_xslparameter directives with
conflicting keys, the latter will always clobber the former.

=head1 NOTES

This package knows how to play nicely with I<Apache::Filter>.

=head1 VERSION

0.3

=head1 DATE

$Date: 2003/09/25 13:42:08 $

=head1 SEE ALSO

L<AxKit>

=head1 LICENSE

Copyright (c) 2002-2003, Aaron Straup Cope. All Rights Reserved.

This is free software, you may use it and distribute it under the same terms as Perl itself.

=cut

